@font-face {
    font-family: 'Fira Sans';
    font-style: normal;
    font-weight: 300;
    src: local('Fira Sans Light'), url("fonts/FiraSans-Light.woff") format('woff');
}
@font-face {
    font-family: 'Fira Sans';
    font-style: normal;
    font-weight: 400;
    src: local('Fira Sans'), url("fonts/FiraSans-Regular.woff") format('woff');
}
@font-face {
    font-family: 'Fira Sans';
    font-style: normal;
    font-weight: 500;
    src: local('Fira Sans Medium'), url("fonts/FiraSans-Medium.woff") format('woff');
}

@font-face {
    font-family: 'Fira Mono';
    font-style: normal;
    font-weight: 400;
    src: local('Fira Mono'), url("fonts/FiraMono-Regular.ttf") format('truetype');
}

@font-face {
    font-family: 'Work Sans';
    font-style: normal;
    font-weight: 500;
    src: local('Work Sans Medium'), url("fonts/WorkSans-Medium.ttf") format('truetype');
}

@font-face {
    font-family: 'Alfa Slab One';
    font-style: normal;
    font-weight: 400;
    src: local('Alfa Slab One'),
        url("fonts/AlfaSlabOne-Regular.woff2") format('woff2'),
        url("fonts/AlfaSlabOne-Regular.woff") format('woff');
}

body {
    margin-top: 2em;
    background-color: white;
    color: #515151;
    font-family: "Fira Sans","Helvetica Neue",Helvetica,Arial,sans-serif;
    font-weight: 300;
    font-size: 25px;
}

pre {
    font-family: "Fira Mono", ui-monospace, SFMono-Regular, Menlo, Monaco, Consolas, "Liberation Mono", "Courier New", monospace;
    font-weight: 400;
}

header > div {
    display: flex;
    flex-direction: column;
    align-items: flex-start;
}

header > div > h1 {
    font-family: "Alfa Slab One", serif;
    color: black;
    font-size: 4rem;
    margin-bottom: 0;
    margin-top: 1rem;
    line-height: 1;
    font-weight: 300;
    letter-spacing: 1px;
}

header > div > h2 {
    color: black;
    font-size: 2rem;
    font-weight: 300;
}

a {
    color: #0b7261;
    text-decoration: underline;
}

a:hover {
    color: #0d8b75;
    text-decoration: underline;
}

main {
    width: 80%;
    max-width: 1200px;
    padding: 0 20px;
    margin-left: auto;
    margin-right: auto;
}

body#idx p {
    margin-top: 2em;
    margin-bottom: 2em;
}

body#idx p.other-platforms-help {
    font-size: 0.6em;
}

.instructions {
    margin-left: auto;
    margin-right: auto;
    text-align: center;
    border-radius: 3px;
    border: 1px solid rgb(204, 204, 204);
    box-shadow: 0px 1px 4px 0px rgb(204, 204, 204);
}

.instructions > * {
    margin-left: auto;
    margin-right: auto;
}

hr {
    border-color: #0b7261;
    margin-top: 2em;
    margin-bottom: 2em;
}

#platform-instructions-unix > p,
#platform-instructions-win32 > p,
#platform-instructions-win64 > p,
#platform-instructions-win-arm64 > p,
#platform-instructions-default > p,
#platform-instructions-unknown > p {
    width: 80%;
}

.rustup-command::before {
    color: black;
    content: " $ ";
	margin-left: 15px;
}

.rustup-command {
    color: black;
    padding: 1rem 1rem 1rem 0;
    height: auto;
    text-align: center;
    border-radius: 3px 0 0 3px;
    box-shadow: inset 0px 0px 20px 0px #f1eeee;
    overflow: hidden;
    font-size: 0.6em;
    white-space: nowrap;
    height: 26px;
    line-height: 26px;
    overflow-x: auto;
}

#platform-instructions-unix div.copy-container,
#platform-instructions-win32 div.copy-container,
#platform-instructions-win64 div.copy-container,
#platform-instructions-win-arm64 div.copy-container,
#platform-instructions-default div.copy-container,
#platform-instructions-unknown div.copy-container {
    display: flex;
    align-items: center;
    width: 90%;
    justify-content: center;
}

#platform-instructions-unix button.copy-button,
#platform-instructions-win32 button.copy-button,
#platform-instructions-win64 button.copy-button,
#platform-instructions-win-arm64 button.copy-button,
#platform-instructions-default button.copy-button,
#platform-instructions-unknown button.copy-button {
    height: 58px;
    margin: 0;
    padding: 0 5px 0 5px;
    border-radius: 0 3px 3px 0;
    border-left-width: 0px;
    border-left-style: solid;
    cursor: pointer;
}

#platform-instructions-unix div.copy-icon,
#platform-instructions-win32 div.copy-icon,
#platform-instructions-win64 div.copy-icon,
#platform-instructions-win-arm64 div.copy-icon,
#platform-instructions-default div.copy-icon,
#platform-instructions-unknown div.copy-icon {
    position: relative;
    width: fit-content;
    height: fit-content;
    top: 27px;
    left: 50%;
    transform: translate(-50%, -50%);
}

#platform-instructions-unix div.copy-button-text,
#platform-instructions-win32 div.copy-button-text,
#platform-instructions-win64 div.copy-button-text,
#platform-instructions-win-arm64 div.copy-button-text,
#platform-instructions-default div.copy-button-text,
#platform-instructions-unknown div.copy-button-text {
    transition: opacity 0.2s ease-in-out;
    opacity: 0;
    font-size: 10px;
    color: #0b7261;
    width: 41px;
    height: 15px;
    position: relative;
    top: 5px;
}

#platform-instructions-win32 a.windows-download,
#platform-instructions-win64 a.windows-download,
#platform-instructions-win-arm64 a.windows-download,
#platform-instructions-default a.windows-download,
#platform-instructions-unknown a.windows-download {
    display: block;
    padding-top: 0.4rem;
    padding-bottom: 0.6rem;
    font-family: "Work Sans", "Fira Sans","Helvetica Neue",Helvetica,Arial,sans-serif;
    font-weight: 500;
    letter-spacing: 0.1rem;
}

/* This is the box that prints navigator.platform, navigator.appVersion values */
#platform-instructions-unknown > div:first-of-type {
    font-size: 16px;
    line-height: 2rem;
}

#help {
    font-size: 20px;
    text-align: center;
}

#about {
    font-size: 16px;
    line-height: 2em;
    text-align: center;
}

#about > img {
    width: 30px;
    height: 30px;
    transform: translateY(11px);
}

#platform-button {
    background-color: #515151;
    color: white;
    margin-left: auto;
    margin-right: auto;
    padding: 1em;
}

.display-none {
    display: none;
}

.display-block {
    display: block;
}

.display-inline {
    display: inline;
}
